-- vymdt.02.2010.10.23.0200
-- Created by: Whys -Open source
--===========================================================================
-- SaveUtils
--===========================================================================
--[[
Credits: killmeplease -- serialize() and deserialize().
Special Thanks: Thalassicus, Afforess.

To use this file, place the following line of code in your modded file's
global scope and set your mod name.

	include( "SaveUtils" ); MY_MOD_NAME = "myMod";

Data caching can be set to one of three states per mod:
0 = no cache, 1 = serialize on save(), 2 = serialize on sync().  Default: 1.

When using serialize on sync(), it is the modders responsibility to call
global function sync() to serialize all cache data, otherwise game data will
not persist between saved games.  Also be aware that another concurrent mod
might choose to inspect your own mod's serialized data at any time.  Ideally,
mods that inspect other mod's data should also inspect the cache to avoid the
possibility of desynchronization.  This may not be possible without
physically combining the mod scripts.

Functionality Examples:
	
	setCacheState( 0 ); -- no cache.
	setCacheState( 0, "otherMod" ); --optional second parameter.

	save( pPlayer, key1, "test" );
	save( pPlayer, key2, 100 );
	save( pPlayer, key3, true );
	save( pPlayer, key1, table, "otherMod" ); --optional fourth parameter.
	save( pPlot,   ...., ...., .... ); --same as above.
	save( pUnit,   ...., ...., .... ); --same as above.

	save(   pPlayer, key1, nil ); --deletes key and value from table.
	delete( pPlayer, key1 );      --deletes key and value from table.
	delete( pPlayer, key1, "otherMod" ); --optional third parameter.

	data = load( pPlayer, key1 );
	data = load( pPlayer, key2 );
	data = load( pPlayer, key3 );
	data = load( pPlayer, key1, "otherMod" ); --optional third parameter.
	data = load( pPlot,   ...., .... ); --same as above.
	data = load( pUnit,   ...., .... ); --same as above.

	data = load( pPlayer, nil );                --mod table for pPlayer.
	data = load( pPlayer, nil, "otherMod" )) ); --mod table for pPlayer.

	data = load( pPlayer, nil, false ); --all data for pPlayer.
	data = load( pPlot,   nil, false ); --all data for pPlot.
	data = load( pUnit,   nil, false ); --all data for pUnit.
]]

-----------------------------------------------------------------------------
-- Save Class
-----------------------------------------------------------------------------
--[[
Allows SetScriptData() and GetScriptData() to behave as a hash-table of
itemized data, while maintaining separation of individual mod data.
]]
-------------------------------------------------
Save = {}; -- Constructor
-------------------------------------------------
function Save:new( o )
	o = o or {};
	setmetatable( o, self );
	self.__index = self;

	--deep assignment of keyed values.
	--required for duplication of table data assigned by key.
	local sub; --allows recursive call to sub-routine.
	sub = function( o )
		for k,v in pairs( o ) do
			self[k] = v;
			if type( v ) == "table" then
				sub( v );
			end
		end
	end

	sub( o );
	return o;
end

-------------------------------------------------
-- Save:set()
-------------------------------------------------
--[[
Sets the given value to the given key for the
given mod.
]]
function Save:set( mod, key, value )
	if key ~= nil then
		self[mod] = self[mod] or {};
		self[mod][key] = value;
		local i = 0;
		for k,v in pairs( self[mod] ) do
			i = i + 1;
		end
		if i == 0 then
			self[mod] = nil;
		end
	end
end
 
-------------------------------------------------
-- Save:get()
-------------------------------------------------
--[[
Returns value for given mod and key.  Returns
self when mod not given.  Returns mod table when
key not given.
]]
function Save:get( mod, key )
	local r = self;
	if mod ~= nil and self[mod] ~= nil then
		r = self[mod];
		if key ~= nil then
			r = self[mod][key];
		end
	end
	return r;
end

-----------------------------------------------------------------------------
--END Save Class
-----------------------------------------------------------------------------

-----------------------------------------------------------------------------
-- Cache Class
-----------------------------------------------------------------------------
--[[
Manages object keys to unserialized object data.
]]
-------------------------------------------------
Cache = {}; -- Constructor
-------------------------------------------------
function Cache:new( o )
	o = o or {};
	setmetatable( o, self );
	self.__index = self;

	--deep assignment of keyed values.
	--required for duplication of table data assigned by key.
	local sub; --allows recursive call to sub-routine.
	sub = function( o )
		for k,v in pairs( o ) do
			self[k] = v;
			if type( v ) == "table" then
				sub( v );
			end
		end
	end

	sub( o );
	return o;
end

-------------------------------------------------
-- Cache:has()
-------------------------------------------------
--[[
Returns boolean state for given target.
]]
function Cache:has( target )
	r = false;
	if self[target] then
		r = true;
	end
	return r;
end

-------------------------------------------------
-- Cache:set()
-------------------------------------------------
--[[
Sets given Save for given target.
]]
function Cache:set( target, save )
	self[target] = save;
end
 
-------------------------------------------------
-- Cache:get()
-------------------------------------------------
--[[
Returns Save for given target.  Returns self when
target not given.
]]
function Cache:get( target )
	local r = self;
	if self:has( target ) then
		r = self[target];
	end
	return r;
end

-----------------------------------------------------------------------------
--END Cache Class
-----------------------------------------------------------------------------

-------------------------------------------------
-- Global Variables.
-------------------------------------------------
DEFAULT_CACHE_STATE = 1;
cacheState = {};
pCache = Cache:new();
MY_MOD_NAME = nil;

-------------------------------------------------
-- setCacheState()
-------------------------------------------------
--[[
Sets caching for given mod to given state.
]]
function setCacheState( state, mod )
	if mod == nil then
		mod = MY_MOD_NAME;
	end
	cacheState[mod] = state;
end

-------------------------------------------------
-- getCacheState()
-------------------------------------------------
--[[
Returns caching state for given mod.
]]
function getCacheState( mod )
	if mod == nil then
		mod = MY_MOD_NAME;
	end
	return cacheState[mod];
end

-------------------------------------------------
-- sync()
-------------------------------------------------
--[[
Serializes all cache data.
]]
function sync()
	for target,save in pairs( pCache ) do
		target:SetScriptData(serialize( save ));
	end
end

-------------------------------------------------
-- save()
-------------------------------------------------
--[[
Saves the given value to the given key for the
given target of the given mod.
]]
function save( target, key, value, mod )
	local pSave = nil;
	local iCache = DEFAULT_CACHE_STATE;

	if mod == nil then
		mod = MY_MOD_NAME;
	end

	pSave = load( target, nil, false );
	pSave:set( mod, key, value );

	if mod ~= nil then
		if getCacheState( mod ) == nil then
			setCacheState( iCache, mod );
		else
			iCache = getCacheState( mod );
		end
	end

	if getCacheState( mod ) <= 1 then
		target:SetScriptData(serialize( pSave ));
	end
	if getCacheState( mod ) >= 1 then
		pCache:set( target, pSave );
	end
end

-------------------------------------------------
-- load()
-------------------------------------------------
--[[
Loads the value for the given key for the given target of the given mod.
Returns entire save table for the given target when mod is false.  Uses
global MY_MOD_NAME If mod not given.
]]
function load( target, key, mod )
	local pSave = nil;
	local iCache = DEFAULT_CACHE_STATE;

	if mod == nil then
		mod = MY_MOD_NAME;
	end
	if mod == false then
		mod = nil;
	end

	if mod ~= nil then
		if getCacheState( mod ) == nil then
			setCacheState( iCache, mod );
		else
			iCache = getCacheState( mod );
		end
	end

	if iCache < 1 then
		pSave = Save:new(deserialize( target:GetScriptData() ));

	elseif iCache > 0 then
		if pCache:has( target ) then
			pSave = pCache:get( target );
		else
			pSave = Save:new(deserialize( target:GetScriptData() ));
			pCache:set( target, pSave );
		end
	end

	return pSave:get( mod, key );
end

-------------------------------------------------
-- delete()
-------------------------------------------------
--[[
Removes both the key and value for the given
target of the given mod with the given key.
Alias for function save( target, key, nil, mod ).
]]
function delete( target, key, mod )
	save( target, key, nil, mod );
end

-------------------------------------------------
-- serialize()
-------------------------------------------------
--[[
Created by: killmeplease.
]]
function serialize( data )
	
	local r = "";
	local sub = function( v )
		if v == nil then
			v = "";
		end
		if v == false or v == true or type( v ) == "number" then
			v = tostring( v );
		else
			v = v:gsub('"', "\[QUOTE\]");
			v = v:gsub('{', "\[LCB\]");
			v = v:gsub('}', "\[RCB\]");
			v = "\""..v.."\""; -- string
		end
		return v;
	end

	if type( data ) ~= "table" then
		r = sub( data );
	else
		r = "{ ";
		local num = 0;
		for k,v in pairs( data ) do
			if num > 0 then
				r = r.." ";
			end
			r = r..k.."=";
			if type( v ) == "table" then
				r = r..serialize( v );
			else
				r = r..sub( v );
			end
			num = num +1;
		end
		r = r.." }";
	end
	return r;
end

-------------------------------------------------
-- deserialize()
-------------------------------------------------
--[[
Created by: killmeplease.
]]
function deserialize( str )
	local tbls = {}; -- nested tables
	local topTbl = {};
	repeat
		topTbl = {}; -- clear variable for top tbl
		-- find first top-level table position in str, its values list and key.
		local strStart, strEnd, topTblKey, topTblVals = str:find("([%w]*)={%s([^{}]*)%s}");
		if topTblKey == nil then
			strStart, strEnd, topTblVals = str:find("{%s([^{}]*)%s}");
		end
		topTblVals = topTblVals or "";
		--print(topTblVals);

		-- parse values:
		-- save string values of top table to a temp table
		local strings = {};
		for vstr in topTblVals:gmatch('"([^"]*)"') do -- match "..." string values
			vstr = vstr:gsub("%[QUOTE%]", '"');
			vstr = vstr:gsub("%[LCB%]", '{');
			vstr = vstr:gsub("%[RCB%]", '}');
			table.insert(strings, vstr);
		end
		-- cut string values from top table string and replace them with "@" symbol
		topTblVals = topTblVals:gsub('"[^"]*"', "@");

		-- parse key=value pairs
		stringNum = 1;
		for k,v in topTblVals:gmatch("(%S+)=(%S+)") do
			-- parse value
			if v == "#" then
				v = tbls[k];
				-- tbls[k] = nil;
			elseif v == "@" then
				v = strings[stringNum];
				stringNum = stringNum +1;
			else
				-- parse simple value, a number or a boolean
				local n = tonumber(v);
				if n ~= nil then
					v = n;
				elseif v == "true" then
					v = true;
				elseif v == "false" then
					v = false;
				end
			end
			-- parse key:
			local n = tonumber(k);
			if n ~= nil then
				k = n;  
			end
			-- add value to the table:
			topTbl[k] = v;
		end
		if topTblKey ~= nil then -- key is nil when it is a base table
			tbls[topTblKey] = topTbl;
			str = str:sub(1, strStart -1)..topTblKey.."=#"..str:sub(strEnd +1);
		else
			str = "#";
		end
		--print(str);
	until str == "#";
	return topTbl;
end

--===========================================================================
--END SaveUtils
--===========================================================================
-- Created by: Whys -Open source